/*
 * read the master list from the ~/.dayplan file. This file is also
 * linked into the daemon program (file_w.c is not).
 *
 *	read_mainlist(retmsg)		Read all files into mainlist.
 *	readfile(list, path)		Read a file into the list.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#ifdef MIPS
#include <sys/types.h>
#include <sys/fcntl.h>
#else
#include <stdlib.h>
#endif
#ifdef ULTRIX
#include <sys/types.h>
#endif
#include <sys/stat.h>
#include <time.h>
#include <errno.h>
#include "conf.h"
#include "version.h"

#define NDAYS		28			/* must agree with cal.h */

#ifdef MIPS
extern char		*getenv();
extern char		*malloc();
#endif
extern char		*mystrdup();
extern time_t		tm_to_time();
extern char		*resolve_tilde();
extern void		destroy_list(), clone_entry(), lockfile();
static int		readfile();

extern char		*progname;		/* argv[0] */
extern struct list	*mainlist;		/* all schedule entries */
extern struct config	config;			/* global configuration data */
struct user		*user;			/* user list for week view */
static int		maxusers;		/* max # of users in u list */
int			nusers;			/* # of users in user list */
char			Print_Spooler[100];	/* print spooling command */


/*
 * read the main list. First, read the main ~/.dayplan file, which also
 * provides the configuration data, including the user list. Next, read
 * the private dayplan file and all user dayplan files. Return a text
 * buffer with error messages; if msg is 0 print themn to stderr. This
 * file can't use create_error_popup because it is used by the daemon too.
 */

BOOL read_mainlist(retmsg)
	char			*retmsg;	/* returned error messages */
{
	struct stat		sbuf;		/* check for directories */
	char			msgbuf[4096];	/* default message buffer */
	char			*msg;		/* where messages are stored */
	char			path[1024];	/* user file name to read */
	int			u;		/* user counter */
	int			error;

	msg = retmsg ? retmsg : msgbuf;
	*msg = 0;
	destroy_list(&mainlist);
	create_list(&mainlist);
							/* main data */
	if (error = readfile(&mainlist, DB_PUB_PATH, 0, TRUE))
		sprintf(msg+strlen(msg),
			"Failed to read master file %s, error %d\n",
			DB_PUB_PATH, error);
	else if (!access(DB_PUB_PATH, F_OK) && access(DB_PUB_PATH, W_OK))
		sprintf(msg+strlen(msg),
		      "Warning: own config and data file %s is not writable\n",
								DB_PUB_PATH);

							/* private data */
	if (error = readfile(&mainlist, DB_PRIV_PATH, 0, FALSE))
		sprintf(msg+strlen(msg),
			"Failed to read private file %s, error %d\n",
			DB_PRIV_PATH, error);
	else if (!access(DB_PRIV_PATH, F_OK) && access(DB_PRIV_PATH, W_OK))
		sprintf(msg+strlen(msg),
			"Warning: own private data file %s is not writable\n",
								DB_PRIV_PATH);

							/* user data */
	for (u=0; u < nusers; u++) {
		if (!user[u].home || !user[u].name) {
			sprintf(msg+strlen(msg),
				"file of user #%d not read, no name or home\n",
								u);
			continue;
		}
		if (!stat(user[u].home, &sbuf) && (sbuf.st_mode & S_IFDIR))
			sprintf(path, "%s/%s", user[u].home, DB_FILE);
		else
			strcpy(path, user[u].home);
		user[u].readonly = !access(path, F_OK) && !!access(path, W_OK);
		user[u].modified = FALSE;
		user[u].time     = get_time();
		if (config.owner_only && !stat(path, &sbuf)
				      && sbuf.st_uid != getuid())
			user[u].readonly = TRUE;

		if ((error = readfile(&mainlist, path, user[u].name, FALSE)) &&
							error != ENOENT)
			sprintf(msg+strlen(msg),
				"Failed to read file %s of user %s, error %d\n",
				user[u].home, user[u].name, error);
	}
	rebuild_repeat_chain(mainlist);

	if (!retmsg)
		fprintf(stderr, "%s", msg);
	return(!*msg);
}


/*
 * read a list from a file. If <list> points to a non-null pointer, the
 * list in the file is merged into the list. As with add_entry(), <list>
 * is a pointer to a pointer to the list because the list might grow and
 * must be re-allocated. If the file is empty, no new list is allocated,
 * and <list> continues to point to a null pointer.
 * If conf is TRUE, read the config parameters too, otherwise ignore them.
 * This is used when other user's .dayplan files are read for the week view.
 * Returns nonzero if the file could not be read.
 */

static int readfile(list, path, name, conf)
	struct list		**list;		/* list to add to */
	char			*path;		/* file to read list from */
	char			*name;		/* user name, 0=own */
	BOOL			conf;		/* if TRUE, config stuff too */
{
	register struct config	*c = &config;
	FILE			*fp;		/* open file */
	BOOL			writable;	/* can we lock the file? */
	char			line[1024];	/* line buffer */
	char			*instring;	/* string at line[2] or [3] */
	register char		*p;		/* line scan pointer */
	register struct entry	*ep = 0;	/* list entry being read */
	int			n;		/* tmp entry # of ep in list */
	struct entry		entry;		/* temp entry before adding */
	struct tm		tm;		/* m/d/y h:m:s to time conv */
	int			tmp[9];		/* temps for sscanf */
	char			sflag,  pflag;	/* more temps for sscanf */
	char			nflag,  nmflag;	/* more temps for sscanf */
	char			nyflag, nwflag;	/* more temps for sscanf */
	int			acol;		/* more temps for sscanf */
	char			**string;	/* string to add to */
	int			len;		/* new length of string */
	int			user_i=nusers;	/* next user to read */
	char			buf1[256], buf2[1024];

	path = resolve_tilde(path);
	writable = !access(path, W_OK);
	if (!(fp = fopen(path, writable ? "r+" : "r")))
		return(errno);
	if (conf)		/* defaults for 't' line */
		guess_tzone();
	if (writable)
		lockfile(fp, TRUE);
	for (;;) {
		if (!fgets(line, 1024, fp))
			break;
		if (!conf && strchr("otelaPpmUu", *line))
			continue;
		instring = line+2;
		switch(*line) {
		  case '\n':
		  case '#':
			continue;

		  case 'V':
			line[strlen(line)-1] = 0;
			sprintf(buf1, "plan %s", VERSION);
			if (strcmp(line, buf1))
				fprintf(stderr,
		"%s: WARNING: file %s is version %s, program version is %s\n",
					progname, path, line, buf1);
		  	break;

		  case 'o':
			c->ndays_in_week  = 7;
			c->sunday_first   = line[2]  != '-';
			c->ampm           = line[3]  != '-';
			c->mmddyy         = line[4]  != '-';
			c->autodel        = line[5]  != '-';
			c->julian         = line[6]  != '-';
			c->weeknum        = line[7]  != '-';
			c->nopast         = line[8]  != '-';
			c->bigwarning     = line[9]  != '-';
			c->weekwarn       = line[10] != '-';
			c->weekuser       = line[11] != '-';
			c->week_bignotime = line[12] != '-';
			c->fullweek       = line[13] != '-';
			c->colorcode_m    = line[14] != '-';
			c->ownonly	  = line[15] != '-';
			sscanf(line+16, "%d %d %d %d %d %d",
						(int *)&c->early_time,
						(int *)&c->late_time,
						(int *)&c->wintimeout,
						(int *)&c->week_minhour,
						(int *)&c->week_maxhour,
						(int *)&c->ndays_in_week);
			if (c->week_minhour >= c->week_maxhour) {
				c->week_minhour = 8;
				c->week_maxhour = 20;
			}
			if (c->ndays_in_week < 1)
				c->ndays_in_week = 7;
			if (c->ndays_in_week > NDAYS)
				c->ndays_in_week = NDAYS;
			break;

		  case 't':
			sscanf(line+2, "%d %d %d %d %d",
						(int *)&c->adjust_time,
						(int *)&c->raw_tzone,
						(int *)&c->dst_flag,
						(int *)&c->dst_begin,
						(int *)&c->dst_end);
			break;

		  case 'e':
			c->ewarn_window = line[2] != '-';
			c->ewarn_mail   = line[3] != '-';
			c->ewarn_exec   = line[4] != '-';
			line[strlen(line)-1] = 0;
			if (c->ewarn_prog)
				free(c->ewarn_prog);
			c->ewarn_prog = 0;
			if (line[5] && line[6])
				c->ewarn_prog = mystrdup(line+6);
			break;

		  case 'l':
			c->lwarn_window = line[2] != '-';
			c->lwarn_mail   = line[3] != '-';
			c->lwarn_exec   = line[4] != '-';
			line[strlen(line)-1] = 0;
			if (c->lwarn_prog)
				free(c->lwarn_prog);
			c->lwarn_prog = 0;
			if (line[5] && line[6])
				c->lwarn_prog = mystrdup(line+6);
			break;

		  case 'a':
			c->alarm_window = line[2] != '-';
			c->alarm_mail   = line[3] != '-';
			c->alarm_exec   = line[4] != '-';
			line[strlen(line)-1] = 0;
			if (c->alarm_prog)
				free(c->alarm_prog);
			c->alarm_prog = 0;
			if (line[5] && line[6])
				c->alarm_prog = mystrdup(line+6);
			break;

		  case 'P':
			c->pr_omit_appts  = line[2] != '-';
			c->pr_omit_priv   = line[3] != '-';
			c->pr_omit_others = line[4] != '-';
			c->pr_mode        = atoi(line+13);
			break;

		  case 'p':
			strncpy(Print_Spooler, (const char *)line + 2,
						sizeof(Print_Spooler));
			Print_Spooler[sizeof(Print_Spooler)-1] = 0;
			Print_Spooler[strlen(Print_Spooler)-1] = 0;
			break;

		  case 'm':
			line[strlen(line)-1] = 0;
			if (line[1] && line[2])
				c->mailer = mystrdup(line+2);
			break;

		  case 'U':
			if (user) {
				for (n=0; n < nusers; n++) {
					if (user[n].name)
						free((char *)user[n].name);
					if (user[n].home)
						free((char *)user[n].home);
				}
				free((char *)user);
				user = 0;
			}
			nusers = user_i = 0;
			sscanf(line+2, "%d", &maxusers);
			if (maxusers) {
				if (!(user = (struct user *)malloc(maxusers *
							sizeof(struct user))))
					fatal("no memory for %d users",
								maxusers);
				for (n=maxusers*sizeof(struct user); n; )
					((char *)user)[--n] = 0;
			}
			break;

		  case 'u':
			if (user_i >= maxusers)
				fatal("too many u stmts in %s", path);
			sscanf(line+2, "%s %s %d %d %d %d\n", buf1, buf2,
					&user[user_i].suspend_w,
					&user[user_i].color,
					&user[user_i].suspend_m,
					&user[user_i].alarm);
			user[user_i].name = mystrdup(buf1);
			user[user_i].home = mystrdup(buf2);
			user_i++;
			nusers++;
			break;

		  case '0':
		  case '1':
		  case '2':
		  case '3':
		  case '4':
		  case '5':
		  case '6':
		  case '7':
		  case '8':
		  case '9':
			clone_entry(&entry, (struct entry *)0);
			sflag = pflag = nflag = nmflag = nyflag = nwflag = 0;
			acol = 0;
			sscanf(line,
	"%d/%d/%d %d:%d:%d %d:%d:%d %d:%d:%d %d:%d:%d %c%c%c%c%c%c---- %d",
				&tm.tm_mon, &tm.tm_mday, &tm.tm_year,
				&tm.tm_hour, &tm.tm_min, &tm.tm_sec,
				tmp+0, tmp+1, tmp+2,
				tmp+3, tmp+4, tmp+5,
				tmp+6, tmp+7, tmp+8,
				&sflag, &pflag, &nflag,
				&nmflag, &nyflag, &nwflag, &acol);
			tm.tm_mon--;
			if (tm.tm_year < 70)
				tm.tm_year += 100;
			if (tm.tm_year > 1900)
				tm.tm_year -= 1900;
			if (entry.notime = tm.tm_hour > 23)
				tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
			entry.time         = tm_to_time(&tm);
			entry.length       = tmp[0]*3600 + tmp[1]*60 + tmp[2];
			entry.early_warn   = tmp[3]*3600 + tmp[4]*60 + tmp[5];
			entry.late_warn    = tmp[6]*3600 + tmp[7]*60 + tmp[8];
			entry.suspended    = sflag  == 'S';
			entry.private      = pflag  == 'P';
			entry.noalarm      = nflag  == 'N';
			entry.hide_in_m    = nmflag == 'M';
			entry.hide_in_y    = nyflag == 'Y';
			entry.hide_in_w    = nwflag == 'W';
			entry.acolor       = acol;
			entry.user	   = mystrdup(name);
			n = add_entry(list, &entry);	/* sets <list> */
			ep = &(*list)->entry[n];
			break;

		  case 'R':
			sscanf(line+2, "%d %d %d %d %d",
				(int *)&ep->rep_every,
				(int *)&ep->rep_last,
				(int *)&ep->rep_weekdays,
				(int *)&ep->rep_days,
				(int *)&ep->rep_yearly);
			break;

		  case 'E':
			for (n=0; n < NEXC; n++)
				if (!ep->except[n])
					break;
			if (n < NEXC) {
				sscanf(line+2, "%d/%d/%d",
					&tm.tm_mon, &tm.tm_mday, &tm.tm_year);
				tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
				tm.tm_mon--;
				if (tm.tm_year < 70)
					tm.tm_year += 100;
				if (tm.tm_year > 1900)
					tm.tm_year -= 1900;
				ep->except[n] = tm_to_time(&tm);
			} else
				fprintf(stderr,
					"too many exception dates for appt");
			break;

		  case 'N':
			string = &ep->note;
			ep->hide_in_w |= line[2] == '-';
			if (line[2] == '-' || line[2] == '=') {
				ep->hide_in_m = ep->hide_in_y = TRUE;
				instring++;
			}
			goto add;

		  case 'M':
			string = &ep->message;
			goto add;

		  case 'S':
			string = &ep->script;
add:			len  = *string ? strlen(*string) : 0;
			len += strlen(instring) + 2;
			if (!(p = malloc(len))) {
				fprintf(stderr, "%s: no memory", progname);
				exit(1);
			}
			if (*string) {
				strcpy(p, *string);
				strcat(p, instring);
				free(*string);
			} else
				strcpy(p, instring);
			*string = p;
			break;
		
		  default:
			fprintf(stderr, "%s: %s: illegal line:\n%s",
							progname, path, line);
		}
	}
	if (writable)
		lockfile(fp, FALSE);
	fclose(fp);
	return(0);
}
